/*
 * evaluate the dc language, from a FILE* or a string
 *
 * Copyright (C) 1994, 1997, 1998, 2000, 2003, 2005, 2006, 2008, 2010, 2012-2017
 * Free Software Foundation, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/* This is the only module which knows about the dc input language */

#include "../config.h"

#include <stdio.h>
#include <signal.h>
#include <string.h>	/* memset */
#include <limits.h> /* UCHAR_MAX */
#ifdef HAVE_UNISTD_H
# include <unistd.h> /* isatty */
#endif
#include "dc.h"
#include "dc-proto.h"

/* an input stream; either a FILE* or a byte-string description: */
typedef struct {
	FILE *fp; /* non-NULL if input stream is a stdio file */
	const char *first, *cur, *last; /* first+last invariant; cur a cursor */
} dc_stream;

typedef enum {DC_FALSE=0, DC_TRUE} dc_bool;

typedef enum {
	DC_OKAY = DC_SUCCESS, /* no further intervention needed for this command */
	DC_EVALTOS,		/* caller needs to eval the string on top of the stack */
	DC_QUIT,		/* quit out of unwind_depth levels of evaluation */
	DC_EOF_ERROR,	/* unexpected end of input; abort current eval */
} dc_status;

static int dc_ibase=10;		/* input base, 2 <= dc_ibase <= DC_IBASE_MAX */
static int dc_obase=10;		/* output base, 2 <= dc_obase */
static int dc_scale=0;		/* scale (see user documentaton) */

/* for Quitting evaluations */
static unsigned int unwind_depth=0;

/* for detecting "infinite" recursion */
static unsigned int recursion_depth=0;

/* if true, active Quit is permitted to exit program */
static dc_bool unwind_exitokay;

/* for handling SIGINT properly */
static volatile sig_atomic_t interrupt_seen=0;

/* forward reference */
static int eval_and_free_str(dc_data *string);


/* read the next input character from supplied stream, or EOF if appropriate */
static int
dc_nextchr(dc_stream *stream)
{
	if (stream->fp)
		return getc(stream->fp);
	if (stream->cur < stream->last  &&  *stream->cur)
		return *stream->cur++;
	return EOF;
}

/* stash "c" back on the front of "stream"; only on character of pushback
 * is guaranteed, and then only if "c" is the most recently read character
 * from "stream" */
static void
dc_unget(dc_stream *stream, int c)
{
	if (c == EOF)
		return; /* we silently ignore this case */
	else if (stream->fp)
		ungetc(c, stream->fp); /* only one-deep guaranteed */
	else if (stream->first < stream->cur)
		--stream->cur; /* assumes "c" was pulled from here */
	/* else assertion error: "c" could not have come from this string */
}

/* hack to allow dc_getnum() to work without it understaning dc_stream */
static dc_stream *cur_input_ptr;

/* passed as an argument to dc_getnum */
static int
num_input_fn(void)
{
	return dc_nextchr(cur_input_ptr);
}


/* step cursor past whitespace and comments */
static const char *
skip_to_non_whitespace(const char *cur, const char *end)
{
	while (cur < end)
		if (*cur==' ' || *cur=='\t' || *cur=='\n')
			++cur;
		else if (*cur == '#')
			while (++cur < end  &&  *cur != '\n')
				continue;
		else
			break;
	return cur;
}

/* notice when an interrupt event happens */
static void
dc_trap_interrupt(int signo)
{
	signal(signo, dc_trap_interrupt);
	interrupt_seen = 1;
}


/* dc_func does the grunt work of figuring out what each input
 * character means; used by both dc_evalstr and dc_evalfile
 */
static dc_status
dc_func(int c, dc_stream *input)
{
	dc_data datum;
	int tmpint;
	int negcmp = 0;

	switch (c){
	case EOF:
		return DC_EOF_ERROR;
	case '_': case '.':
	case '0': case '1': case '2': case '3':
	case '4': case '5': case '6': case '7':
	case '8': case '9': case 'A': case 'B':
	case 'C': case 'D': case 'E': case 'F':
		dc_unget(input, c);
		cur_input_ptr = input;
		dc_push(dc_getnum(num_input_fn, dc_ibase, &tmpint));
		dc_unget(input, tmpint);
		break;
	case ' ': case '\t': case '\n':
		/* standard command separators */
		break;
	case '#':	/* comment; skip remainder of current line */
		do {
			c = dc_nextchr(input);
		} while (c!=EOF && c!='\n');
		break;

	case '+':	/* add top two stack elements */
		dc_binop(dc_add, dc_scale);
		break;
	case '-':	/* subtract top two stack elements */
		dc_binop(dc_sub, dc_scale);
		break;
	case '*':	/* multiply top two stack elements */
		dc_binop(dc_mul, dc_scale);
		break;
	case '/':	/* divide top two stack elements */
		dc_binop(dc_div, dc_scale);
		break;
	case '%':
		/* take the remainder from division of the top two stack elements */
		dc_binop(dc_rem, dc_scale);
		break;
	case '~':
		/* Do division on the top two stack elements.
		 * Return the quotient as next-to-top of stack
		 * and the remainder as top-of-stack.
		 */
		dc_binop2(dc_divrem, dc_scale);
		break;
	case '^':	/* exponientiation of the top two stack elements */
		dc_binop(dc_exp, dc_scale);
		break;
	case '|':
		/* Consider the top three elements of the stack as (base, exp, mod),
		 * where mod is top-of-stack, exp is next-to-top, and base is
		 * just below that. Mod must be non-zero, exp must be non-negative,
		 * and all three must be integers. Push the result of raising
		 * base to the exp power, reduced modulo mod. If we had base in
		 * register b, exp in register e, and mod in register m then this
		 * is conceptually equivalent to "lble^lm%", but it is implemented
		 * in a more efficient manner, and can handle arbitrarily large
		 * values for exp.
		 */
		dc_triop(dc_modexp, dc_scale);
		break;
	case 'v':	/* replace top of stack with its square root */
		dc_monop(dc_sqrt, dc_scale);
		break;

	case 'c':	/* clear whole stack */
		dc_clear_stack();
		break;
	case 'd':	/* duplicate the datum on the top of stack */
		if (dc_top_of_stack(&datum) == DC_SUCCESS)
			dc_push(dc_dup(datum));
		break;
	case 'f':	/* print list of all stack items */
		dc_printall(dc_obase);
		break;
	case 'n':
		/* print the value popped off of top-of-stack;
		 * do not add a trailing newline
		 */
		if (dc_pop(&datum) == DC_SUCCESS)
			dc_print(datum, dc_obase, DC_NONL, DC_TOSS);
		break;
	case 'p':
		/* print the datum on the top of stack (without popping it),
		 * with a trailing newline
		 */
		if (dc_top_of_stack(&datum) == DC_SUCCESS)
			dc_print(datum, dc_obase, DC_WITHNL, DC_KEEP);
		break;
	case 'P':
		/* Pop the value off the top of a stack.  If it is
		 * a number, dump out the integer portion of its
		 * absolute value as a "base UCHAR_MAX+1" byte stream;
		 * if it is a string, just print it.
		 * In either case, do not append a trailing newline.
		 */
		if (dc_pop(&datum) == DC_SUCCESS){
			if (datum.dc_type == DC_NUMBER)
				dc_dump_num(datum.v.number, DC_TOSS);
			else if (datum.dc_type == DC_STRING)
				dc_out_str(datum.v.string, DC_TOSS);
			else
				dc_garbage("at top of stack", -1);
		}
		dc_flush_stdout(0);
		break;
	case 'r':	/* rotate (swap) the top two elements on the stack */
		dc_stack_rotate(2);
		break;
	case 'R':
		/* pop a value off of the evaluation stack;
		 * rotate the top remaining stack elements that many
		 * places forward (negative numbers mean rotate backward).
		 */
		if (dc_int_from_TOS(&tmpint, 0) != DC_SUCCESS) {
			dc_warn("rotation count must be numeric");
			break;
		}
		dc_discard_TOS();
		dc_stack_rotate(tmpint);
		break;

	case '[':	/* read to balancing ']' into a dc_str */
		/*cheating here and looking inside the *input structure*/
		if (input->fp) {
			datum = dc_readstring(input->fp, '[', ']');
			if ( ! ferror(input->fp) )
				dc_push(datum);
		} else {
			const char *start = input->cur;
			int count = 1;
			size_t len;
			for (; input->cur < input->last && count>0; ++input->cur) {
				if (*input->cur == ']')
					--count;
				else if (*input->cur == '[')
					++count;
			}
			len = input->cur - start;
			dc_push(dc_makestring(start, (count==0 ? len-1 : len)));
		}
		break;
	case ']':	/* common error, give gentler message*/
		dc_warn("unbalanced ']'");
		break;
	case 'a':	/* Convert top of stack to an ascii character. */
		if (dc_pop(&datum) == DC_SUCCESS){
			char tmpc = '?';
			static int want_warn = 1;
			if (want_warn) {
				dc_warn("warning: 'a' command is deprecated"
					"\n\t(contact <bug-dc@gnu.org> if you actually use it)");
				want_warn = 0;
			}
			if (datum.dc_type == DC_NUMBER){
				dc_push(datum);
				dc_push(dc_int2data(1+UCHAR_MAX));
				dc_binop(dc_rem, dc_scale);
				if (dc_pop(&datum) != DC_SUCCESS){
					/*impossible*/
				}
				tmpc = (char) dc_num2int(datum.v.number, '?');
				dc_free_num(&datum.v.number);
			}else if (datum.dc_type == DC_STRING){
				tmpc = *dc_str2charp(datum.v.string);
				dc_free_str(&datum.v.string);
			}else{
				dc_garbage("at top of stack", -1);
			}
			dc_push(dc_makestring(&tmpc, 1));
		}
		break;

	case 'i':	/* set input base to value on top of stack */
		if (dc_int_from_TOS(&tmpint, -1) != DC_SUCCESS
				||  tmpint < 2  ||  DC_IBASE_MAX < tmpint) {
			fprintf(stderr,
			"%s: input base must be a number between 2 and %d (inclusive)\n",
					progname, DC_IBASE_MAX);
			break;
		}
		dc_discard_TOS();
		dc_ibase = tmpint;
		break;
	case 'k':	/* set scale to value on top of stack */
		if (dc_int_from_TOS(&tmpint, -1) != DC_SUCCESS  ||  tmpint < 0) {
			dc_warn("scale must be a nonnegative number");
			break;
		}
		dc_discard_TOS();
		dc_scale = tmpint;
		break;
	case 'o':	/* set output base to value on top of stack */
		if (dc_int_from_TOS(&tmpint, -1) != DC_SUCCESS  ||  tmpint < 2) {
			dc_warn("output base must be a number greater than 1");
			break;
		}
		dc_discard_TOS();
		dc_obase = tmpint;
		break;
	case 'I':	/* push the current input base onto the stack */
		dc_push(dc_int2data(dc_ibase));
		break;
	case 'K':	/* push the current scale onto the stack */
		dc_push(dc_int2data(dc_scale));
		break;
	case 'O':	/* push the current output base onto the stack */
		dc_push(dc_int2data(dc_obase));
		break;
	case 'z':	/* push the current stack depth onto the top of stack */
		dc_push(dc_int2data(dc_tell_stackdepth()));
		break;
	case 'X':	/* replace the number on top-of-stack with its scale factor */
		if (dc_pop(&datum) == DC_SUCCESS){
			tmpint = 0;
			if (datum.dc_type == DC_NUMBER)
				tmpint = dc_tell_scale(datum.v.number, DC_TOSS);
			dc_push(dc_int2data(tmpint));
		}
		break;
	case 'Z':	/* replace the datum on the top-of-stack with its length */
		if (dc_pop(&datum) == DC_SUCCESS)
			dc_push(dc_int2data(dc_tell_length(datum, DC_TOSS)));
		break;

	case '!':	/* read to newline and call system() on resulting string */
		/* read to newline and call system() on resulting string,
		 * except if part of digraph !< != !>, in which case
		 * run a negated comparison
		 */
		c = dc_nextchr(input);
		if ( ! (c == '<' || c == '=' || c == '>') ){
			dc_unget(input, c);
			/*cheating here and looking inside the *input structure*/
			if (input->fp) {
				datum = dc_readstring(input->fp, '\n', '\n');
				if (ferror(input->fp))
					return DC_EOF_ERROR;
				(void)dc_system(dc_str2charp(datum.v.string));
				dc_free_str(&datum.v.string);
			} else {
				input->cur = dc_system(input->cur);
			}
			break;
		}
		/*else, not "system" call after all, but a negated comparison:*/
		negcmp = !negcmp;
		/*@fallthrough@*/
	case '<':
	case '=':
	case '>':
		/* if relation holds for top two stack elements, which must
		 * both be numeric, then eval the register named by dc_nextchr()
		 */
		tmpint = ! dc_cmpop(c, !negcmp);
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (tmpint  ||  dc_register_get(c, &datum) != DC_SUCCESS)
			break;
		dc_push(datum);
		return DC_EVALTOS;
	case 'x':	/* eval the datum popped from top of stack */
		return DC_EVALTOS;
	case '?':	/* read a line from standard-input and eval it */
		datum = dc_readstring(stdin, '\n', '\n');
		if (ferror(stdin)){
			dc_free_str(&datum.v.string);
			return DC_EOF_ERROR;
		}
		dc_push(datum);
		return DC_EVALTOS;
	case 'q':	/* quit two levels of evaluation, possibly exiting program */
		unwind_depth = 1; /* the return below is the first level of returns */
		unwind_exitokay = DC_TRUE;
		return DC_QUIT;
	case 'Q':
		/* quit out of top-of-stack nested evals; pops value from stack;
		 * does not exit program (stops short if necessary)
		 */
		if (dc_int_from_TOS(&tmpint, -1) != DC_SUCCESS  ||  tmpint < 1) {
			dc_warn("Q command requires a number >= 1");
			unwind_depth = 0; /*paranoia*/
			break;
		}
		dc_discard_TOS();
		unwind_depth = tmpint - 1; /*return below is first unwind*/
		unwind_exitokay = DC_FALSE;
		return DC_QUIT;

	case 'l':
		/* "load" -- push value on top of register stack named by dc_nextchr()
		 * onto top of evaluation stack; does not modify the register stack
		 */
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (dc_register_get(c, &datum) == DC_SUCCESS)
			dc_push(datum);
		break;
	case 's':
		/* "store" -- replace top of register stack named by dc_nextchr()
		 * with the value popped from the top of the evaluation stack
		 */
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (dc_pop(&datum) == DC_SUCCESS)
			dc_register_set(c, datum);
		break;
	case 'L':
		/* pop a value off of register stack named by dc_nextchr()
		 * and push it onto the evaluation stack
		 */
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (dc_register_pop(c, &datum) == DC_SUCCESS)
			dc_push(datum);
		break;
	case 'S':
		/* pop a value off of the evaluation stack and push it
		 * onto the register stack named by dc_nextchr()
		 */
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (dc_pop(&datum) == DC_SUCCESS)
			dc_register_push(c, datum);
		break;
	case ':':	/* store into array */
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (dc_int_from_TOS(&tmpint, -1) != DC_SUCCESS  ||  tmpint < 0) {
			dc_warn("array index must be a nonnegative integer");
		} else {
			dc_data value;
			dc_discard_TOS();
			if (dc_pop(&value) == DC_SUCCESS)
				dc_array_set(c, tmpint, value);
			else
				dc_push(dc_int2data(tmpint)); /*restore stack*/
		}
		break;
	case ';':	/* retreive from array */
		if ((c = dc_nextchr(input)) == EOF)
			return DC_EOF_ERROR;
		if (dc_int_from_TOS(&tmpint, -1) != DC_SUCCESS  ||  tmpint < 0) {
			dc_warn("array index must be a nonnegative integer");
			break;
		}
		dc_discard_TOS();
		dc_push(dc_array_get(c, tmpint));
		break;

	default:	/* What did that user mean? */
		fprintf(stderr, "%s: ", progname);
		dc_show_id(stdout, c, " unimplemented\n");
		break;
	}
	return DC_OKAY;
}


/* takes a string and evals it */
static int
evalstr(dc_data *string)
{
	unsigned int tail_depth = 1; /* how much tail recursion is active */
	dc_stream input;
	dc_data evalstr;

	if (string->dc_type != DC_STRING){
		dc_warn("eval called with non-string argument");
		return DC_OKAY;
	}
	input.fp = NULL;
	input.cur = input.first = dc_str2charp(string->v.string);
	input.last = input.first + dc_strlen(string->v.string);
	while (input.cur < input.last  &&  interrupt_seen == 0){
		switch (dc_func(dc_nextchr(&input), &input)){
		case DC_OKAY:
			break;
		case DC_EVALTOS:
			/*skip trailing whitespace to assist tail-recursion detection*/
			input.cur = skip_to_non_whitespace(input.cur, input.last);
			if (dc_pop(&evalstr) == DC_SUCCESS){
				if (evalstr.dc_type == DC_NUMBER){
					dc_push(evalstr);
				}else if (evalstr.dc_type != DC_STRING){
					dc_garbage("at top of stack", -1);
				}else if (input.cur == input.last){
					/*handle tail recursion*/
					dc_free_str(&string->v.string);
					*string = evalstr;
					input.cur = input.first = dc_str2charp(string->v.string);
					input.last = input.first + dc_strlen(string->v.string);
					++tail_depth;
				}else if (eval_and_free_str(&evalstr) == DC_QUIT){
					if (unwind_depth > 0){
						--unwind_depth;
						return DC_QUIT;
					}
					return DC_OKAY;
				}
			}
			break;
		case DC_QUIT:
			if (unwind_depth >= tail_depth){
				/*unwind what we own, then propagate to caller*/
				unwind_depth -= tail_depth;
				return DC_QUIT;
			}
			/*adjust tail recursion accounting and continue*/
			tail_depth -= unwind_depth;
			break;
		case DC_EOF_ERROR:
			if (ferror(stdin)){
				fprintf(stderr, "%s: ", progname);
				perror("error reading stdin");
				return DC_FAIL;
			}
			dc_warn("unexpected EOS");
			return DC_OKAY;
		}
	}
	return DC_OKAY;
}

/* takes a string and evals it; frees the string when done */
/* Wrapper around evalstr to avoid duplicating the free call
 * at all possible return points.
 */
static int
eval_and_free_str(dc_data *string)
{
	dc_status status;

	++recursion_depth;
	if (max_recursion_depth && max_recursion_depth <= recursion_depth)
		interrupt_seen = 2;
	status = evalstr(string);
	if (string->dc_type == DC_STRING)
		dc_free_str(&string->v.string);
	--recursion_depth;
	return status;
}

/* wrapper around evalstr, to handle top-level QUIT requests correctly*/
int
dc_evalstr(dc_data *string)
{
	interrupt_seen = recursion_depth = 0;
	switch (evalstr(string)) {
	case DC_OKAY:
		return DC_SUCCESS;
	case DC_QUIT:
		if (unwind_exitokay)
			return DC_EXIT_REQEST;
		return DC_SUCCESS;
	default:
		return DC_FAIL;
	}
}


/* This is the main function of the whole DC program.
 * Reads the file described by fp, calls dc_func to do
 * the dirty work, and takes care of dc_func's shortcomings.
 */
int
dc_evalfile(FILE *fp)
{
	typedef void (*handler_t)(int);
	handler_t sigint_handler = dc_trap_interrupt;
	handler_t sigint_default = signal(SIGINT, SIG_IGN);
	dc_data datum;
	dc_stream input;
	int c;

	/* Signals are awkward: we want to allow interactive users
	 * to break out of long running macros, but otherwise we
	 * prefer that SIGINT not be given any special treatment.
	 * Sometimes "no special treatment" means to continue to
	 * *ignore* the signal, but usually it means to kill the program.
	 */
	signal(SIGINT, sigint_default);
#ifdef HAVE_UNISTD_H
	/* don't trap SIGINT unless we can tell that we are reading from a tty */
	if ( ! isatty(fileno(fp)) )
		sigint_handler = sigint_default;
#endif

	memset(&input, 0, sizeof input); /*paranoia*/
	input.fp = fp;
	for (;;) {
		if ((c = getc(fp)) == EOF)
			goto reset_and_exit_success;
		if (sigint_handler == dc_trap_interrupt)
			signal(SIGINT, sigint_handler);
		switch (dc_func(c, &input)){
		case DC_OKAY:
			break;
		case DC_EVALTOS:
			if (dc_pop(&datum) == DC_SUCCESS){
				if (datum.dc_type == DC_NUMBER){
					dc_push(datum);
				}else if (datum.dc_type == DC_STRING){
					interrupt_seen = recursion_depth = 0;
					if (eval_and_free_str(&datum) == DC_QUIT){
						if (unwind_exitokay == DC_TRUE)
							goto reset_and_exit_quit;
						/*Q command argument exceeded string execution depth*/
					}
				}else{
					dc_garbage("at top of stack", -1);
				}
			}
			break;
		case DC_QUIT:
			if (unwind_exitokay == DC_TRUE)
				goto reset_and_exit_quit;
			/*Q command argument exceeded string execution depth*/
			break;
		case DC_EOF_ERROR:
			if (ferror(fp))
				goto error_fail;
			dc_warn("unexpected EOF");
			goto reset_and_exit_fail;
		}
		if (ferror(fp))
			goto error_fail;

		if (interrupt_seen == 2)
			fprintf(stderr, "\nmax_recursion_depth exceeded; evaluation interrupted!\n");
		else if (interrupt_seen)
			fprintf(stderr, "\nInterrupt!\n");
		interrupt_seen = 0;
		if (sigint_handler == dc_trap_interrupt)
			signal(SIGINT, sigint_default);
	}

error_fail:
	signal(SIGINT, sigint_default);
	fprintf(stderr, "%s: ", progname);
	perror("error reading input");
	return DC_FAIL;
reset_and_exit_quit:
reset_and_exit_fail:
	signal(SIGINT, sigint_default);
	return DC_FAIL;
reset_and_exit_success:
	signal(SIGINT, sigint_default);
	return DC_SUCCESS;
}


/*
 * Local Variables:
 * mode: C
 * tab-width: 4
 * End:
 * vi: set ts=4 :
 */
